@@ -53,7 +53,7 @@ soup_str_case_equal (gconstpointer v1,
 /* Base64 utils (straight from camel-mime-utils.c) */
 #define d(x)
 
-static char *base64_alphabet =
+static const char *base64_alphabet =
 	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 /* 
@@ -217,11 +217,10 @@ char *
 soup_base64_encode (const char *text, int len)
 {
         unsigned char *out;
-        int state = 0, outlen;
-        unsigned int save = 0;
+        int state = 0, outlen,  save = 0;
         
         out = g_malloc (len * 4 / 3 + 5);
-        outlen = soup_base64_encode_close (text, 
+        outlen = soup_base64_encode_close ((const guchar *)text,
 					   len, 
 					   FALSE,
 					   out, 
@@ -318,15 +317,17 @@ char *
 soup_base64_decode (const char   *text,
 		    int          *out_len)
 {
-	char *ret;
-	int inlen, state = 0, save = 0;
+	guchar *ret;
+	int inlen, state = 0;
+	unsigned int save = 0;
 
 	inlen = strlen (text);
 	ret = g_malloc0 (inlen);
 
-	*out_len = soup_base64_decode_step (text, inlen, ret, &state, &save);
+	*out_len = soup_base64_decode_step ((const guchar *)text, inlen,
+					    ret, &state, &save);
 
-	return ret; 
+	return (char *)ret; 
 }
 
 typedef struct {
@@ -392,3 +393,98 @@ soup_signal_connect_once (gpointer instance, const char *detailed_signal,
 						    closure, FALSE);
 	return ssod->signal_id;
 }
+
+/**
+ * soup_add_io_watch:
+ * @async_context: the #GMainContext to dispatch the I/O watch in, or
+ * %NULL for the default context
+ * @chan: the #GIOChannel to watch
+ * @condition: the condition to watch for
+ * @function: the callback to invoke when @condition occurs
+ * @data: user data to pass to @function
+ *
+ * Adds an I/O watch as with g_io_add_watch(), but using the given
+ * @async_context.
+ *
+ * Return value: a #GSource, which can be removed from @async_context
+ * with g_source_destroy().
+ **/
+GSource *
+soup_add_io_watch (GMainContext *async_context,
+		   GIOChannel *chan, GIOCondition condition,
+		   GIOFunc function, gpointer data)
+{
+	GSource *watch = g_io_create_watch (chan, condition);
+	g_source_set_callback (watch, (GSourceFunc) function, data, NULL);
+	g_source_attach (watch, async_context);
+	g_source_unref (watch);
+	return watch;
+}
+
+/**
+ * soup_add_idle:
+ * @async_context: the #GMainContext to dispatch the idle event in, or
+ * %NULL for the default context
+ * @function: the callback to invoke at idle time
+ * @data: user data to pass to @function
+ *
+ * Adds an idle event as with g_idle_add(), but using the given
+ * @async_context.
+ *
+ * Return value: a #GSource, which can be removed from @async_context
+ * with g_source_destroy().
+ **/
+GSource *
+soup_add_idle (GMainContext *async_context,
+	       GSourceFunc function, gpointer data)
+{
+	GSource *source = g_idle_source_new ();
+	g_source_set_callback (source, function, data, NULL);
+	g_source_attach (source, async_context);
+	g_source_unref (source);
+	return source;
+}
+
+/**
+ * soup_add_timeout:
+ * @async_context: the #GMainContext to dispatch the timeout in, or
+ * %NULL for the default context
+ * @interval: the timeout interval, in milliseconds
+ * @function: the callback to invoke at timeout time
+ * @data: user data to pass to @function
+ *
+ * Adds a timeout as with g_timeout_add(), but using the given
+ * @async_context.
+ *
+ * Return value: a #GSource, which can be removed from @async_context
+ * with g_source_destroy().
+ **/
+GSource *
+soup_add_timeout (GMainContext *async_context,
+		  guint interval,
+		  GSourceFunc function, gpointer data)
+{
+	GSource *source = g_timeout_source_new (interval);
+	g_source_set_callback (source, function, data, NULL);
+	g_source_attach (source, async_context);
+	g_source_unref (source);
+	return source;
+}
+
+/**
+ * soup_xml_real_node:
+ * @node: an %xmlNodePtr
+ *
+ * Finds the first "real" node (ie, not a comment or whitespace) at or
+ * after @node at its level in the tree.
+ *
+ * Return: a node, or %NULL
+ **/
+xmlNode *
+soup_xml_real_node (xmlNode *node)
+{
+	while (node && (node->type == XML_COMMENT_NODE ||
+			xmlIsBlankNode (node)))
+		node = node->next;
+	return node;
+}
